__author__ = 'Dmitriy Korsakov'

API_KEY_ID = None

API_SECRET_KEY = None

API_VERSION = "v1beta0"

envId = None

view = "tree"

debug_mode = False

colored_output = True

API_SCHEME = 'https'

API_HOST = 'my.scalr.net'

SIGNATURE_VERSION = 'V1-HMAC-SHA256'



__author__ = 'Dmitriy Korsakov'
__doc__ = 'Cost Centrer management'


from scalrctl import commands


name = "cost-center"
enabled = True


def callback(*args, **kwargs):
    """
    print('in cost-center module')
    print(args)
    print(kwargs)
    """
    pass


class CostCenterList(commands.SubCommand):
    name = "list"
    route = "/{envId}/cost-centers/"
    method = "get"
    enabled = True


class CostCenterRetrieve(commands.SubCommand):
    name = "retrieve"
    route = "/{envId}/cost-centers/{costCenterId}/"
    method = "get"
    enabled = True



__author__ = 'Dmitriy Korsakov'

import os
import json

import settings


class MetaSpec(object):

    specs = None

    data = {
        "account": {
            "os": {
                    "descr" : "Supported operating systems",
                    "subcommands": {
                        "list": ("/os/", "get"),
                        "retrieve": ("/os/{osId}/", "get"),
                        },
            },

            "event": {
                    "descr" : "Manage events available in the Scalr Account",
                    "subcommands": {
                        "list": ("/events/", "get"),
                        "retrieve": ("/events/{eventId}/", "get"),
                        "create": ("/events/", "post"),
                        "delete": ("/events/{eventId}/", "delete"),
                        },
                    },

            "image": {
                    "descr" : "Image management",
                    "subcommands": {
                        "change-attributes": ("/images/{imageId}/", "patch"),
                        "copy": ("/images/{imageId}/actions/copy/", "post"),
                        "register": ("/images/", "post"),
                        "retrieve": ("/images/{imageId}/", "get"),
                        "list": ("/images/", "get"),
                        "delete": ("/images/{imageId}/", "delete"),
                        },
                    },

            "role-orchestration-rule": {
                    "descr" : "Orchestration Rule management",
                    "subcommands": {
                        "list": ("/roles/{roleId}/orchestration-rules/", "get"),
                        "retrieve": ("/roles/{roleId}/orchestration-rules/{orchestrationRuleId}/", "get"),
                        "create": ("/roles/{roleId}/orchestration-rules/", "post"),
                        "delete": ("/roles/{roleId}/orchestration-rules/{orchestrationRuleId}/", "delete"),
                        },
                    },

            "role-global-variables": {
                    "descr" : "Manage global variables for roles",
                    "subcommands": {
                        "list": ("/roles/{roleId}/global-variables/", "get"),
                        "retrieve": ("/roles/{roleId}/global-variables/{globalVariableName}/", "get"),
                        "update": ("/roles/{roleId}/global-variables/{globalVariableName}/", "patch"),
                        "create": ("/roles/{roleId}/global-variables/", "post"),
                        "delete": ("/roles/{roleId}/global-variables/{globalVariableName}/", "delete"),
                        },
                    },

            "role-image": {
                    "descr" : "Manage images in roles",
                    "subcommands": {
                        "list": ("/roles/{roleId}/images/", "get"),
                        "retrieve": ("/roles/{roleId}/images/{imageId}/", "get"),
                        "update": ("/roles/{roleId}/images/", "post"),
                        "delete": ("/roles/{roleId}/images/{imageId}/", "delete"),
                        "replace": ("/roles/{roleId}/images/{imageId}/actions/replace/", "post"),
                        },
                    },

            "role": {
                    "descr" : "Role management",
                    "subcommands": {
                        "change-attributes": ("/roles/{roleId}/", "patch"),
                        "list": ("/roles/", "get"),
                        "retrieve": ("/roles/{roleId}/", "get"),
                        "create": ("/roles/", "post"),
                        "delete": ("/roles/{roleId}/", "delete"),
                        },
                    },

        }
    }


    def __init__(self, *specs):
        self.specs = specs

    def __repr__(self):
        return str([spec_dict["basePath"] for spec_dict in self.specs])


    @classmethod
    def lookup(cls, base_path=None, api_levels=None):
        if not base_path:
            base_path = os.path.expanduser(os.environ.get("SCALRCLI_HOME", "~/.scalr"))

        if not api_levels:
            api_levels = ["user", "account", "admin"]

        data = []

        for api_level in api_levels:
            path = os.path.join(base_path, "%s.json" % api_level)
            if os.path.exists(path):
                spec_dict = json.load(open(path, "r"))
                data.append(spec_dict)

        return cls(*data)

    def get_route(self, command_name, subcommand_name, api_level="user"):
        return self.data[api_level][command_name]["subcommands"][subcommand_name][0]

    def get_http_method(self, command_name, subcommand_name, api_level="user"):
        return self.data[api_level][command_name]["subcommands"][subcommand_name][1]

    def list_http_methods(self, route):
        methods = []
        return methods

    def get_command_name(self, route, method):
        name = []
        return name

    def get_command_object(self, command_name, subcommand_name, api_level="user"):
        # THIS!
        obj = None
        return obj

    def list_options(self, command_name, subcommand_name, api_level="user"):
        pass


    def _get_spec_dict(self, api_level="user"):
        for spec in self.specs:
            if api_level in spec["basePath"]:
                return spec

    def _list_cmd_aliases(self, api_level="user"):
        """return route aliases"""
        level_data = self.data[api_level]
        return [alias for alias in level_data if alias]

    def _get_cmd_descr(self, command_name, api_level="user"):
        "returns help for command e.g. OS help"
        level_data = self.data[api_level]
        cmd_data = level_data[command_name]
        return cmd_data["descr"]

    def _get_subcmd_descr(self, command_name, subcommand_name, api_level="user"):
        "returns help for command e.g. OS help"
        level_data = self.data[api_level]
        cmd_data = level_data[command_name]
        return cmd_data["descr"]

    def _list_subcmd_aliases(self, command_name, api_level="user"):
        """returns aliases for subcommands (route+http_method)"""
        return self.data[api_level][command_name]["subcommands"].keys()


class Spec(object):

    document = None
    method = None
    route = None

    def __init__(self, document, route, method):
        self.document = document
        self.route = route
        self.method = method

    @property
    def _route_data(self):
        return self.document["paths"][self.route]

    @property
    def _data(self):
        return self._route_data[self.method]

    @property
    def _result_descr(self):
        responses = self._data['responses']
        if '200' in responses:
            ok200 = responses['200']
            if 'schema' in ok200:
                schema = ok200['schema']
                if '$ref' in schema:
                    reference = schema['$ref']
                    return self.lookup(reference)

    @property
    def description(self):
        return self._data['description'] if 'description' in self._data else ''

    @property
    def method_params(self):
        params = []
        if "parameters" in self._data:
            for parameter in self._data['parameters']:
                params.append(parameter)
        return params

    @property
    def route_params(self):
        return self._route_data['parameters']

    def returns_iterable(self):
        if "GET" != self.method.upper():
            return False

        object_properties = self._result_descr["properties"]
        data_structure = object_properties["data"]
        if "type" in data_structure:
            responce_type = data_structure["type"]
            if "array" == responce_type:
                return True
        return False

    @property
    def filters(self):
        if self.returns_iterable():
            response_ref = self._result_descr["properties"]["data"]["items"]["$ref"]
            response_descr = self.lookup(response_ref)
            if "x-filterable" in response_descr:
                filters = response_descr["x-filterable"]
                return filters
        return []

    def lookup(self, refstr):
        """
        Returns document section
        Example: #/definitions/Image returns Image defenition section
        """
        if refstr.startswith("#"):
            paths = refstr.split("/")[1:]
            result = self.document
            for path in paths:
                if path not in result:
                    return
                result = result[path]
            return result

    def get_column_names(self):
        fields = []
        data = self._result_descr["properties"]["data"]
        # response_ref = data["items"]["$ref"] if "items" in data else data["$ref"] # [ST-54]
        response_ref = data["items"]["$ref"]
        response_descr = self.lookup(response_ref)
        for k,v in response_descr["properties"].items():
            if "$ref" not in v:
                fields.append(k)
        return sorted(fields)

    def __repr__(self):
        return 'Spec("%s", "%s")' % (self.route, self.method)

def get_raw_spec(api_level="user"):
    api_level = api_level or "user"  #XXX: Ugly,  SubCommand class needs to be changed first
    path = os.path.join(os.path.expanduser(os.environ.get("SCALRCLI_HOME", "~/.scalr")), "%s.json" % api_level)
    if os.path.exists(path):
        return json.load(open(path, "r"))

def get_spec_url(api_level="user"):
    api_level = api_level or "user"  #XXX: Ugly,  SubCommand class needs to be changed first
    return "{0}://{1}/api/{2}.{3}.yml".format(settings.API_SCHEME, settings.API_HOST, api_level, settings.API_VERSION)


__author__ = 'Dmitriy Korsakov'
__doc__ = 'Image management'

from scalrctl import commands, settings
import click
import json
import sys
import inspect


name = "image"
enabled = True


def callback(*args, **kwargs):
    """
    print('in image module')
    print(args)
    print(kwargs)
    """
    pass


class Image(commands.SubCommand):
    pass


class ChangeImageAttrs(Image):
    name = "change-attributes"
    route = "/{envId}/images/{imageId}/"
    method = "patch"
    enabled = True
    mutable_body_parts = ["name"]
    prompt_for = ["imageId"]


    def modify_options(self, options):
        options = super(ChangeImageAttrs, self).modify_options(options)
        return options


class CopyImage(Image):
    name = "copy"
    route = "/{envId}/images/{imageId}/actions/copy/"
    method = "post"
    enabled = True


class DeleteImage(Image):
    name = "delete"
    route = "/{envId}/images/{imageId}/"
    method = "delete"
    enabled = True


class ListImages(Image):
    name = "list"
    route = "/{envId}/images/"
    method = "get"
    enabled = True


class RegisterImage(Image):
    name = "register"
    route = "/{envId}/images/"
    method = "post"
    enabled = True

    def pre(self, *args, **kwargs):
        #XXX: this code does not work yet!
        if 'imageId' not in kwargs:
            kwargs["image"] = click.termui.prompt("Image object JSON")
        return args, kwargs


class RetrieveImage(Image):
    name = "retrieve"
    route = "/{envId}/images/{imageId}/"
    method = "get"
    enabled = True
    prompt_for = ["imageId"]











__author__ = 'Dmitriy Korsakov'
__doc__ = 'Orchestration Rule management'


from scalrctl import commands


name = "role-orchestration-rule"
enabled = True


def callback(*args, **kwargs):
    """
    print('in role-orchestration-rule module')
    print(args)
    print(kwargs)
    """
    pass


class OrchestrationRule(commands.SubCommand):
    pass


class ListOrchestrationRules(OrchestrationRule):
    name = "list"
    route = "/{envId}/roles/{roleId}/orchestration-rules/"
    method = "get"
    enabled = True


class RetrieveOrchestrationRule(OrchestrationRule):
    name = "retrieve"
    route = "/{envId}/roles/{roleId}/orchestration-rules/{orchestrationRuleId}/"
    method = "get"
    enabled = True


class CreateOrchestrationRule(OrchestrationRule):
    name = "create"
    route = "/{envId}/roles/{roleId}/orchestration-rules/"
    method = "post"
    enabled = True


class ChangeOrchestrationRuleAttrs(OrchestrationRule):
    name = "change-attributes"
    route = "/{envId}/roles/{roleId}/orchestration-rules/{orchestrationRuleId}/"
    method = "patch"
    enabled = True


class DeleteOrchestrationRule(OrchestrationRule):
    name = "delete"
    route = "/{envId}/roles/{roleId}/orchestration-rules/{orchestrationRuleId}/"
    method = "delete"
    enabled = True



__author__ = 'Dmitriy Korsakov'

import os
import sys
import json
import shutil
import inspect

import yaml
import click
import requests

import commands
import settings
import spec

PROGNAME = "scalr-ctl"
DEFAULT_PROFILE = "default"
CMD_FOLDER = os.path.abspath(os.path.join(os.path.dirname(__file__), 'commands'))
CONFIG_FOLDER = os.path.expanduser(os.environ.get("SCALRCLI_HOME", "~/.scalr"))
CONFIG_PATH = os.path.join(CONFIG_FOLDER, "%s.yaml" % os.environ.get("SCALRCLI_PROFILE", DEFAULT_PROFILE))

SWAGGER_USER_NOUPDATE_TRIGGER = ".noupdate.user"
SWAGGER_USER_FILE = "user.yaml"
SWAGGER_USER_PATH = os.path.join(CONFIG_FOLDER, SWAGGER_USER_FILE)
SWAGGER_USER_JSONSPEC_FILE = SWAGGER_USER_FILE.split(".")[0] + ".json"
SWAGGER_USER_JSONSPEC_PATH = os.path.join(CONFIG_FOLDER, SWAGGER_USER_JSONSPEC_FILE)

SWAGGER_ACCOUNT_NOUPDATE_TRIGGER = ".noupdate.account"
SWAGGER_ACCOUNT_FILE = "account.yaml"
SWAGGER_ACCOUNT_PATH = os.path.join(CONFIG_FOLDER, SWAGGER_ACCOUNT_FILE)
SWAGGER_ACCOUNT_JSONSPEC_FILE = SWAGGER_ACCOUNT_FILE.split(".")[0] + ".json"
SWAGGER_ACCOUNT_JSONSPEC_PATH = os.path.join(CONFIG_FOLDER, SWAGGER_ACCOUNT_JSONSPEC_FILE)

AUTOCOMPLETE_FNAME = "path.bash.inc"
AUTOCOMPLETE_CONTENT = "_%s_COMPLETE=source %s" % (PROGNAME.upper().replace("-", "_"), PROGNAME)
AUTOCOMPLETE_PATH = os.path.join(os.path.expanduser(CONFIG_FOLDER), AUTOCOMPLETE_FNAME)


def setup_bash_complete():
    if "nt" == os.name: # Click currently only supports completion for Bash.
        return

    bashrc_path = os.path.expanduser("~/.bashrc")
    bashprofile_path = os.path.expanduser("~/.bash_profile")
    startup_path = bashprofile_path if os.path.exists(bashprofile_path) else bashrc_path
    startup_path = click.prompt("Enter path to an rc file to update, or leave blank to use", default=startup_path, err=True)
    if not os.path.exists(startup_path):
        click.echo("%s not found." % startup_path)
        return
    startupfile_content = open(startup_path, "r").read()

    if AUTOCOMPLETE_PATH not in startupfile_content:
        confirmed = click.confirm("Modify profile to update your $PATH and enable bash completion?", default=True, err=True)

        if confirmed:
            with open(AUTOCOMPLETE_PATH, "w") as fp:
                fp.write(AUTOCOMPLETE_CONTENT)

            backup_path = startup_path + ".backup"
            click.echo("Backing up [%s] to [%s]." % (startup_path, backup_path))
            shutil.copy(startup_path, backup_path)

            #source_line = "source '%s'" % AUTOCOMPLETE_PATH #XXX: for some reason this seized to work
            source_line = 'eval "$(%s)"' % AUTOCOMPLETE_CONTENT

            comment = "# The next line enables bash completion for %s." % PROGNAME
            newline = "" if startupfile_content.endswith("\n") else "\n"
            add = "%s%s\n%s" % (newline, comment, source_line)

            # Handling pip install --user and PATH
            local_binpath = os.path.join(os.path.expanduser("~/.local/bin/"), PROGNAME)
            if os.path.exists(local_binpath):
                add += "\nasias %s=%s" % (PROGNAME, local_binpath)

            with open(startup_path, "a") as afp:
                afp.write(add)

            click.echo("Start a new shell for the changes to take effect.")


def configure(profile=None):
    """
    Configure command-line client.
    Creates new profile in configuration directory
    and downloads spec file.
    :param profile: Profile name
    """
    confpath = os.path.join(CONFIG_FOLDER, "%s.yaml" % profile) if profile else CONFIG_PATH
    data = {}

    if os.path.exists(confpath):
        old_data = yaml.load(open(confpath, "r"))
        data.update(old_data)

    click.echo("Configuring %s:" % confpath)

    for obj in dir(settings):
        if not obj.startswith("__"):
            default_value = getattr(settings, obj)
            if isinstance(default_value, bool):
                data[obj] = click.confirm(obj, default=getattr(settings, obj))
            elif not default_value or type(default_value) in (int, str):
                data[obj] = str(click.prompt(obj, default=getattr(settings, obj))).strip()

    if not os.path.exists(CONFIG_FOLDER):
        os.makedirs(CONFIG_FOLDER)

    raw = yaml.dump(data, default_flow_style=False, default_style='')
    with open(confpath, 'w') as fp:
        fp.write(raw)

    click.echo()
    click.echo("New config saved:")
    click.echo()
    click.echo(open(confpath, "r").read())

    for setting, value in data.items():
        setattr(settings, setting, value)

    update()
    setup_bash_complete()


def update():
    """
    Downloads yaml spec and converts it to JSON
    Both files are stored in configuration directory.
    """
    text = None
    user_trigger_file = os.path.join(CONFIG_FOLDER, SWAGGER_USER_NOUPDATE_TRIGGER)

    user_url = spec.get_spec_url(api_level="user")
    user_dst = os.path.join(CONFIG_FOLDER, SWAGGER_USER_FILE)

    if user_url and not os.path.exists(user_trigger_file):
        click.echo("Trying to get new UserAPI Spec from %s" % user_url)
        r = requests.get(user_url)

        old = None

        if os.path.exists(user_dst):
            with open(user_dst, "r") as fp:
                old = fp.read()

        text = r.text

        if text == old:
            click.echo("UserAPI Spec is already up-to-date.")
        elif text:
            with open(user_dst, "w") as fp:
                fp.write(text)
            click.echo("UserAPI UserSpec successfully updated.")

    if text or os.path.exists(user_dst):
        struct = yaml.load(text or open(user_dst).read())
        json.dump(struct, open(SWAGGER_USER_JSONSPEC_PATH, "w"))


    # Fetch AccountAPI spec and convert to JSON
    text = None
    account_trigger_file = os.path.join(CONFIG_FOLDER, SWAGGER_ACCOUNT_NOUPDATE_TRIGGER)
    account_url = spec.get_spec_url(api_level="account")
    account_dst = os.path.join(CONFIG_FOLDER, SWAGGER_ACCOUNT_FILE)

    if account_url and not os.path.exists(account_trigger_file):
        click.echo("Trying to get new AccountAPI Spec from %s" % account_url)
        r = requests.get(account_url)

        old = None

        if os.path.exists(account_dst):
            with open(account_dst, "r") as fp:
                old = fp.read()

        text = r.text

        if text == old:
            click.echo("AccountAPI Spec is already up-to-date.")
        elif text:
            with open(account_dst, "w") as fp:
                fp.write(text)
            click.echo("AccountAPI Spec successfully updated.")

    if text or os.path.exists(account_dst):
        struct = yaml.load(text or open(account_dst).read())
        json.dump(struct, open(SWAGGER_ACCOUNT_JSONSPEC_PATH, "w"))


class HelpBuilder(object):
    document = None

    def __init__(self, document):
        #XXX: move methods to spec module
        self.document = document

    def list_paths(self):
        return self.document["paths"].keys()

    def list_http_methods(self, path):
        l = self.document["paths"][path].keys()
        if "parameters" in l:
            l.remove("parameters")
        return l

    def get_method_description(self, path, method="get"):
        return self.document["paths"][path][method]['description']

    def get_body_type_params(self, path, method="get"):
        params = []
        m = self.document["paths"][path][method]
        if "parameters" in m:
            for parameter in m['parameters']:
                params.append(parameter)
        return params

    def get_path_type_params(self, path):
        params = []
        paths = self.document["paths"].keys()
        d = self.document["paths"][path]
        if "parameters" in d:
            for parameter in d['parameters']:
                params.append(parameter)
        return params

    def get_params(self, path, method="get"):
        result = self.get_path_type_params(path)
        if method.upper() in ("GET", "DELETE"):
            body_params = self.get_body_type_params(path, method)
            result += body_params
        return result

    def returns_iterable(self, path):
        responces = self.document["paths"][path]["get"]['responses']
        if '200' in responces:
            ok200 = responces['200']
            if 'schema' in ok200:
                schema = ok200['schema']
                if '$ref' in schema:
                    reference = schema['$ref']
                    object_key = reference.split("/")[-1]
                    object_descr = self.document["definitions"][object_key]
                    object_properties = object_descr["properties"]
                    data_structure = object_properties["data"]
                    if "type" in data_structure:
                        responce_type = data_structure["type"]
                        if "array" == responce_type:
                            return True
        return False


def list_module_filenames():
    files = os.listdir(CMD_FOLDER)
    """
    account_files = os.listdir(os.path.join(CMD_FOLDER, "account"))
    print account_files
    files += account_files
    """
    l = [fname for fname in files if fname.endswith('.py') and not fname.startswith("_")]
    return l



class MyCLI(click.Group):

    _modules = None

    def __init__(self, name=None, commands=None, **attrs):
        click.Group.__init__(self, name, commands, **attrs)
        self._modules = {}
        self._init()
        self.metaspec = spec.MetaSpec.lookup()


    def get_help_builder(self, api_level="user"):
        spec_dict = self.metaspec._get_spec_dict(api_level=api_level)
        return HelpBuilder(spec_dict)


    def _list_module_objects(self):
        objects = [module for module in self._modules.values() if module.enabled]
        #print "MODULE OBJECTS: %s" % objects
        return objects


    def _list_subcommands(self, command_name):
        objects = []
        for name, obj in inspect.getmembers(self._modules[command_name]):
            if inspect.isclass(obj) and hasattr(obj, 'enabled') and getattr(obj, 'enabled'):
                subcommand = obj()
                if isinstance(subcommand, commands.SubCommand):
                    objects.append(subcommand)
        #print "SUBCOMMANDS: %s", objects
        return objects


    def _list_options(self, route, method, subcommand_name, api_level="user"):  #XXX: subcommand_name
        help_builder = self.get_help_builder(api_level=api_level)
        params = help_builder.get_params(route, method)
        options = []

        debug = click.Option(('--debug/--no-debug', 'debug'), default=False, help="Print debug messages")
        options.append(debug)

        for param in params:
            option = click.Option(("--%s" % param['name'], param['name']), required=param['required'], help=param["description"])
            options.append(option)

        if method.upper() == 'GET':
            raw = click.Option(('--raw', 'transformation'), is_flag=True, flag_value='raw', default=False, help="Print raw response")
            tree = click.Option(('--tree', 'transformation'), is_flag=True, flag_value='tree', default=True, help="Print response as a colored tree")
            nocolor = click.Option(('--nocolor', 'nocolor'), is_flag=True, default=False, help="Use colors")
            options += [raw, tree, nocolor]
            if subcommand_name != "retrieve": # [ST-54]
                table = click.Option(('--table', 'transformation'), is_flag=True, flag_value='table', default=False, help="Print response as a colored table")
                options.append(table)

            if help_builder.returns_iterable(route):
                maxrez = click.Option(("--maxresults", "maxResults"), type=int, required=False, help="Maximum number of records. Example: --maxresults=2")
                options.append(maxrez)

                pagenum = click.Option(("--pagenumber", "pageNum"), type=int, required=False, help="Current page number. Example: --pagenumber=3")
                options.append(pagenum)

                filthelp = "Apply filters. Example: type=ebs,size=8. "
                spc = spec.Spec(spec.get_raw_spec(api_level=api_level), route, method)
                if spc.filters:
                    filters = sorted(spc.filters)
                    filthelp += "Available filters: %s." % ", ".join(filters)
                    filters = click.Option(("--filters", "filters"), required=False, help=filthelp)
                    options.append(filters)

                columnshelp = "Filter columns in table view [--table required]. Example: NAME,SIZE,SCOPE. "
                available_columns = spc.get_column_names()
                columnshelp +=  "Available columns: %s." % ", ".join(available_columns)
                columns = click.Option(("--columns", "columns"), required=False, help=columnshelp)
                options.append(columns)

        if method.upper() in ('PATCH','POST'):
            stdin_help = "Ask for input instead of opening default text editor"
            stdin = click.Option(("--stdin", "stdin"), is_flag=True, default=False, help=stdin_help)
            options.append(stdin)

        return options


    def _init(self):
        for name in list_module_filenames():
            try:
                if sys.version_info[0] == 2:
                    name = name.encode('ascii', 'replace')
                mod = __import__('scalrctl.commands.' + name[:-3], None, None, ['enabled'])
                if hasattr(mod, "name"):
                    self._modules[mod.name] = mod
            except ImportError:
                raise  # pass
        #print self._modules


    def list_commands(self, ctx):
        rv = [module.name for module in self._list_module_objects()]
        rv += ["configure", "update", "account"]
        rv.sort()
        return rv


    def get_command(self, ctx, name):

        if name == "account":
            account_help = "All AccountAPI commands"  # TODO: HelpStr
            account_group = click.Group("account", callback=lambda: None, help=account_help)

            #print self.metaspec._list_subcmd_aliases(command_name="os", api_level="account")
            #print self.metaspec._list_cmd_aliases(api_level="account")

            for command_name in self.metaspec._list_cmd_aliases(api_level="account"):
                command_descr = self.metaspec._get_cmd_descr(command_name=command_name, api_level="account")
                grp = click.Group(command_name, callback=lambda: None, help=command_descr)

                for subcommand_name in self.metaspec._list_subcmd_aliases(command_name=command_name, api_level="account"):

                    route = self.metaspec.get_route(command_name, subcommand_name, api_level="account")
                    method = self.metaspec.get_http_method(command_name, subcommand_name, api_level="account")
                    subcommand = commands.SubCommand()
                    subcommand.name = subcommand_name
                    subcommand.route = route
                    subcommand.method = method
                    subcommand.api_level = "account"
                    options = self._list_options(route, method, subcommand_name, api_level="account")
                    options = subcommand.modify_options(options)
                    acc_spec = spec.Spec(spec.get_raw_spec(api_level="account"), route, method)
                    subcommand_descr = acc_spec.description
                    cmd = click.Command(subcommand_name, params=options, callback=subcommand.run, help=subcommand_descr)
                    grp.add_command(cmd)

                account_group.add_command(grp)

            """
            for subcommand in self._list_subcommands(name):
                if subcommand.route in self.hb.list_paths() \
                        and subcommand.method in self.hb.list_http_methods(subcommand.route):
                    options = self._list_options(subcommand)

                    options = subcommand.modify_options(options)

                    spc = spec.Spec(spec.get_raw_spec(), subcommand.route, subcommand.method)
                    cmd = click.Command(subcommand.name, params=options, callback=subcommand.run, help=spc.description)
                    group.add_command(cmd)
            """


            return account_group

        elif name == "configure":
            configure_help = "Set configuration options in interactive mode"
            profile_argument = click.Argument(("profile",), required=False) # [ST-30]
            configure_cmd = click.Command("configure", callback=configure, help=configure_help, params=[profile_argument,])
            return configure_cmd

        elif name == "update":
            update_help = "Fetch new API specification if available."
            update_cmd = click.Command("update", callback=update, help=update_help)
            return update_cmd

        elif name not in self._modules:
            raise click.ClickException("No such command: %s" % name)

        group = click.Group(name, callback=self._modules[name].callback, help=self._modules[name].__doc__)

        hb = self.get_help_builder(api_level="user")
        subcommands = self._list_subcommands(name)  #TODO: BROKEN CODE! see scalr-ctl os list
        routes = hb.list_paths()
        #print routes
        for subcommand in subcommands:
            #print subcommand.route
            if subcommand.route in routes \
                    and subcommand.method in hb.list_http_methods(subcommand.route):
                options = self._list_options(route=subcommand.route, method=subcommand.method, subcommand_name=subcommand.name, api_level="user")

                options = subcommand.modify_options(options)

                spc = spec.Spec(spec.get_raw_spec(api_level="user"), subcommand.route, subcommand.method)
                cmd = click.Command(subcommand.name, params=options, callback=subcommand.run, help=spc.description)
                group.add_command(cmd)

        return group

def account():
    # debug code
    pass

def apply_settings(data):
    for key, value in data.items():
        if hasattr(settings, key):
            setattr(settings, key, value)


if not os.path.exists(CONFIG_FOLDER):
    os.makedirs(CONFIG_FOLDER)

if os.path.exists(CONFIG_PATH):
    apply_settings(yaml.load(open(CONFIG_PATH, "r")))

if not os.path.exists(SWAGGER_USER_PATH) or not os.path.exists(SWAGGER_USER_JSONSPEC_PATH):
    update() # [ST-53]


@click.command(cls=MyCLI)
@click.version_option()
@click.pass_context
@click.option('--key_id', help="API key ID")
@click.option('--secret_key', help="API secret key")
def cli(ctx, key_id, secret_key, *args, **kvargs):
    """Scalr-ctl is a command-line interface to your Scalr account"""
    if key_id:
        settings.API_KEY_ID = str(key_id)
    if secret_key:
        settings.API_SECRET_KEY = str(secret_key)
    elif settings.API_KEY_ID and settings.API_KEY_ID.strip() and not settings.API_SECRET_KEY: # [ST-21]
        if ctx.invoked_subcommand not in ("configure", "update"):
            raw = click.prompt(text="API SECRET KEY", hide_input=True)
            settings.API_SECRET_KEY = str(raw)


if __name__ == '__main__':
    cli()



__author__ = 'Dmitriy Korsakov'
__doc__ = 'Role management'


from scalrctl import commands


name = "role"
enabled = True


def callback(*args, **kwargs):
    """
    print('in role module')
    print(args)
    print(kwargs)
    """
    pass


class Role(commands.SubCommand):
    pass


class ChangeRoleAttrs(Role):
    name = "change-attributes"
    route = "/{envId}/roles/{roleId}/"
    method = "patch"
    enabled = True
    prompt_for = ["roleId"]


class CreateRole(Role):
    name = "create"
    route = "/{envId}/roles/"
    method = "post"
    enabled = True


class DeleteRole(Role):
    name = "delete"
    route = "/{envId}/roles/{roleId}/"
    method = "delete"
    enabled = True


class ListRoles(Role):
    name = "list"
    route = "/{envId}/roles/"
    method = "get"
    enabled = True


class RetrieveRole(Role):
    name = "retrieve"
    route = "/{envId}/roles/{roleId}/"
    method = "get"
    enabled = True






__author__ = 'Dmitriy Korsakov'
__doc__ = 'Farm management'


from scalrctl import commands


name = "farm"
enabled = True


def callback(*args, **kwargs):
    """
    print('in farm module')
    print(args)
    print(kwargs)
    """
    pass


class Farm(commands.SubCommand):
    pass


class ListFarms(Farm):
    name = "list"
    route = "/{envId}/farms/"
    method = "get"
    enabled = True


class CreateFarm(Farm):
    name = "create"
    route = "/{envId}/farms/"
    method = "post"
    enabled = True


class RetrieveFarm(Farm):
    name = "retrieve"
    route = "/{envId}/farms/{farmId}/"
    method = "get"
    enabled = True


class ChangeFarmAttrs(Farm):
    name = "change-attributes"
    route = "/{envId}/farms/{farmId}/"
    method = "patch"
    enabled = True


class DeleteFarm(Farm):
    name = "delete"
    route = "/{envId}/farms/{farmId}/"
    method = "delete"
    enabled = True


class LaunchFarm(Farm):
    name = "launch"
    route = "/{envId}/farms/{farmId}/actions/launch/"
    method = "post"
    enabled = True


class TerminateFarm(Farm):
    name = "terminate"
    route = "/{envId}/farms/{farmId}/actions/terminate/"
    method = "post"
    enabled = True


__author__ = 'Dmitriy Korsakov'
__doc__ = 'View projects'


from scalrctl import commands


name = "project"
enabled = True


def callback(*args, **kwargs):
    """
    print('in project module')
    print(args)
    print(kwargs)
    """
    pass


class ProjectList(commands.SubCommand):
    name = "list"
    route = "/{envId}/projects/"
    method = "get"
    enabled = True


class CreateProject(commands.SubCommand):
    name = "create"
    route = "/{envId}/projects/"
    method = "post"
    enabled = True


class ProjectRetrieve(commands.SubCommand):
    name = "retrieve"
    route = "/{envId}/projects/{projectId}/"
    method = "get"
    enabled = True



__author__ = 'Dmitriy Korsakov'
__doc__ =  'Manage events available in the Scalr Environment'


from scalrctl import commands


name = "event"
enabled = True


def callback(*args, **kwargs):
    """
    print('in env-event module')
    print(args)
    print(kwargs)
    """
    pass


class EnvEvent(commands.SubCommand):
    pass


class ListEvents(EnvEvent):
    name = "list"
    route = "/{envId}/events/"
    method = "get"
    enabled = True


class RetrieveEvent(EnvEvent):
    name = "retrieve"
    route = "/{envId}/events/{eventId}/"
    method = "get"
    enabled = True


class CreateEvent(EnvEvent):
    name = "create"
    route = "/{envId}/events/"
    method = "post"
    enabled = True


class ChangeEventAttrs(EnvEvent):
    name = "change-attributes"
    route = "/{envId}/events/{eventId}/"
    method = "patch"
    enabled = True


class DeleteEvent(EnvEvent):
    name = "delete"
    route = "/{envId}/events/{eventId}/"
    method = "delete"
    enabled = True



__author__ = 'Dmitriy Korsakov'
__doc__ = 'Manage images in roles'


from scalrctl import commands


name = "role-image"
enabled = True


def callback(*args, **kwargs):
    """
    print('in image module')
    print(args)
    print(kwargs)
    """
    pass


class RoleImage(commands.SubCommand):
    pass


class UpdateImage(RoleImage):
    name = "update"
    route = "/{envId}/roles/{roleId}/images/"
    method = "post"
    enabled = True


class ListImages(RoleImage):
    name = "list"
    route = "/{envId}/roles/{roleId}/images/"
    method = "get"
    enabled = True


class ImageDetails(RoleImage):
    name = "retrieve"
    route = "/{envId}/roles/{roleId}/images/{imageId}/"
    method = "get"
    enabled = True


class DeleteImage(RoleImage):
    name = "delete"
    route = "/{envId}/roles/{roleId}/images/{imageId}/"
    method = "delete"
    enabled = True


class ReplaceImage(RoleImage):
    name = "replace"
    route = '/{envId}/roles/{roleId}/images/{imageId}/actions/replace/'
    method = "post"
    enabled = True



__author__ = 'Dmitriy Korsakov'
__doc__ = 'FarmRole management'


from scalrctl import commands


name = "farm-role"
enabled = True


def callback(*args, **kwargs):
    """
    print('in farm-role module')
    print(args)
    print(kwargs)
    """
    pass


class FarmRole(commands.SubCommand):
    pass


class RetrieveFarmRole(FarmRole):
    name = "retrieve"
    route = "/{envId}/farm-roles/{farmRoleId}/"
    method = "get"
    enabled = True


class ListFarmRoles(FarmRole):
    name = "list"
    route = "/{envId}/farms/{farmId}/farm-roles/"
    method = "get"
    enabled = True


class CreateFarmRole(FarmRole):
    name = "create"
    route = "/{envId}/farms/{farmId}/farm-roles/"
    method = "post"
    enabled = True


class ChangeFarmRoleAttrs(FarmRole):
    name = "change-attributes"
    route = "/{envId}/farm-roles/{farmRoleId}/"
    method = "patch"
    enabled = True


class DeleteFarmRole(FarmRole):
    name = "delete"
    route = "/{envId}/farm-roles/{farmRoleId}/"
    method = "delete"
    enabled = True


class RetrievePlacementConfiguration(FarmRole):
    name = "retrieve-placement-configuration"
    route = "/{envId}/farm-roles/{farmRoleId}/placement/"
    method = "get"
    enabled = True


class UpdatePlacementConfiguration(FarmRole):
    name = "change-placement-configuration"
    route = "/{envId}/farm-roles/{farmRoleId}/placement/"
    method = "patch"
    enabled = True


class RetrieveInstanceConfiguration(FarmRole):
    name = "retrieve-instance-configuration"
    route = "/{envId}/farm-roles/{farmRoleId}/instance/"
    method = "get"
    enabled = True


class UpdateInstanceConfiguration(FarmRole):
    name = "change-instance-configuration"
    route = "/{envId}/farm-roles/{farmRoleId}/instance/"
    method = "patch"
    enabled = True


class RetrieveScalingConfiguration(FarmRole):
    name = "retrieve-scaling-configuration"
    route = "/{envId}/farm-roles/{farmRoleId}/scaling/"
    method = "get"
    enabled = True


class UpdateScalingConfiguration(FarmRole):
    name = "change-scaling-configuration"
    route = "/{envId}/farm-roles/{farmRoleId}/scaling/"
    method = "patch"
    enabled = True


__author__ = 'Dmitriy Korsakov'
__doc__ = 'Manage global variables for roles'


import sys
from scalrctl import commands


name = "role-global-variables"
enabled = True


def callback(*args, **kwargs):
    """
    print('in role-global-variables module')
    print(args)
    print(kwargs)
    """
    pass


class GV(commands.SubCommand):
    route = "/{envId}/roles/{roleId}/global-variables/{globalVariableName}/"


class DeleteRoleGlobalVariable(GV):
    name = "delete"
    method = "delete"
    enabled = True


class ListRoleGlobalVariables(GV):
    route = "/{envId}/roles/{roleId}/global-variables/"
    name = "list"
    method = "get"
    enabled = True


class NewRoleGV(GV):
    route = "/{envId}/roles/{roleId}/global-variables/"
    name = "create"
    method = "post"
    enabled = True


class RetrieveRoleGlobalVariable(GV):
    name = "retrieve"
    method = "get"
    enabled = True


class UpdateRoleGlobalVariable(GV):
    name = "update"
    method = "patch"
    enabled = True
    #object_reference = "#/definitions/GlobalVariable"
    prompt_for = ["roleId", "globalVariableName"]


__author__ = 'Dmitriy Korsakov'

import os
import time
import hmac
import hashlib
import binascii
import urlparse
import urllib
import json
import yaml

import requests
from scalrctl import spec

CONFIG_FOLDER = os.path.expanduser(os.environ.get("SCALRCLI_HOME", "~/.scalr"))
SWAGGER_USER_FILE = "user.yaml"
SWAGGER_USER_PATH = os.path.join(CONFIG_FOLDER, SWAGGER_USER_FILE)
SWAGGER_USER_JSONSPEC_FILE = SWAGGER_USER_FILE.split(".")[0] + ".json"
SWAGGER_USER_JSONSPEC_PATH = os.path.join(CONFIG_FOLDER, SWAGGER_USER_JSONSPEC_FILE)


class BaseConnection(object):

    def __init__(self, api_host, api_scheme, api_key_id, api_secret_key, api_version, signature_version, env_id):
        self.api_host = api_host
        self.api_scheme = api_scheme
        self.api_key_id = api_key_id
        self.api_secret_key = api_secret_key
        self.api_version = api_version
        self.signature_version = signature_version
        self.env_id = env_id


    def _get_signature(self, method, request_uri, payload, body, time_iso8601):
        query_string = urllib.urlencode(sorted(payload.items())) if payload else ''
        string_to_sign = "%s\n%s\n%s\n%s\n%s" % (method.upper(), time_iso8601, request_uri, query_string, body)
        digest = hmac.new(self.api_secret_key, string_to_sign, hashlib.sha256).digest()
        return binascii.b2a_base64(digest).strip()


    def request(self, method, request_uri, payload=None, data=None):
        body = json.dumps(yaml.safe_load(data)) if data else '' #XXX
        time_iso8601 = time.strftime("%Y-%m-%dT%H:%M:%S.000Z", time.gmtime())
        signature = self._get_signature(method, request_uri, payload, body, time_iso8601)

        headers = dict()
        headers['Content-Type'] = 'application/json; charset=utf-8'
        headers['X-Scalr-Key-Id'] =  self.api_key_id
        headers['X-Scalr-Date'] = time_iso8601
        headers['X-Scalr-Signature'] = "%s %s" % (self.signature_version, signature)

        url = urlparse.urlunsplit((self.api_scheme, self.api_host, request_uri, '', ''))
        r = requests.request(method.lower(), url, data=body, params=payload, headers=headers)
        return json.loads(r.text)


class Connection(BaseConnection):

    _spec = None

    def __init__(self, api_host, api_scheme, api_key_id, api_secret_key, api_version, signature_version, env_id):
        super(Connection, self).__init__(
            api_host, api_scheme, api_key_id, api_secret_key, api_version, signature_version, env_id
        )
        self._spec = self._get_spec()


    def _get_spec_url(self, api_level="user"):
        return "{0}://{1}/api/{2}.{3}.yml".format(self.api_scheme, self.api_host, api_level, self.api_version)


    def _get_spec(self):
        """
        Downloads yaml spec and converts it to JSON
        JSON file is stored in configuration directory.
        """
        if self._spec:
            return

        if os.path.exists(SWAGGER_USER_JSONSPEC_PATH):
            struct = json.load(open(SWAGGER_USER_JSONSPEC_PATH))
        else:
            r = requests.get(self._get_spec_url())
            struct = yaml.load(r.text)
            json.dump(struct, open(SWAGGER_USER_JSONSPEC_PATH, "w"))

        return struct


    def list_methods(self):
        """
        #TODO: automatically download spec
        """
        l = []
        for route in self._spec["paths"]:
            for method in self._spec["paths"][route].keys():
                if method != "parameters":
                    #print '("%s", "%s") : ("%s", "%s")' % (method, route, self._get_route_alias(route), method)
                    l.append((str(route), str(method)))
        return l


    def _get_route_alias(self, route):
        parts = route.split('/')

        if len(parts) < 2:
            return

        alias = ''

        for part in parts:
            if part and not part.startswith('{'):
                alias += "_%s" % part.replace("-", "_")

        return alias[1:]


if __name__ == "__main__":

    import settings

    conn = Connection(
        api_host=settings.API_HOST,
        api_scheme=settings.API_SCHEME,
        api_key_id=settings.API_KEY_ID or os.environ.get("SCALR_API_KEY_ID"),
        api_secret_key=settings.API_SECRET_KEY or os.environ.get("SCALR_API_SECRET_KEY"),
        api_version=settings.API_VERSION,
        signature_version=settings.SIGNATURE_VERSION,
        env_id=settings.envId
    )

    #print conn.request("get", "/api/user/v1beta0/os/")
    for m in conn.list_methods():
        print m


__author__ = 'Dmitriy Korsakov'
__doc__ = 'Current Role categories'

from scalrctl import commands


name = "role-category"
enabled = True


def callback(*args, **kwargs):
    """
    print('in role_category module')
    print(args)
    print(kwargs)
    """
    pass


class RoleCategory(commands.SubCommand):
    pass


class RoleCategoryRetrieve(RoleCategory):
    name = "retrieve"
    route = "/{envId}/role-categories/{roleCategoryId}/"
    method = "get"
    enabled = True

class RoleCategoryList(RoleCategory):
    name = "list"
    route = "/{envId}/role-categories/"
    method = "get"
    enabled = True



import sys
import time
import hmac
import hashlib
import binascii
import urlparse
import urllib
import json
import yaml

import click
import requests

from scalrctl import settings


"""
import logging

# these two lines enable debugging at httplib level (requests->urllib3->httplib)
# you will see the REQUEST, including HEADERS and DATA, and RESPONSE with HEADERS but without DATA.
# the only thing missing will be the response.body which is not logged.

import httplib
httplib.HTTPConnection.debuglevel = 1


logging.basicConfig() # you need to initialize logging, otherwise you will not see anything from requests
logging.getLogger().setLevel(logging.DEBUG)
requests_log = logging.getLogger("requests.packages.urllib3")
requests_log.setLevel(logging.DEBUG)
requests_log.propagate = True
"""


try:
    requests.packages.urllib3.disable_warnings()
except:
    pass

def request(method, request_uri, payload=None, data=None):
    """
    #XXX:temporary!
    if not payload:
        payload = {}
    payload["id"] = "0059d6b9-6def-4b5b-3a36-256ed5c80001"
    print "payload:", payload
    """
    scheme = settings.API_SCHEME
    api_host = settings.API_HOST
    time_iso8601 = time.strftime("%Y-%m-%dT%H:%M:%S.000Z", time.gmtime())
    try:
        r = None
        query_string = urllib.urlencode(sorted(payload.items())) if payload else ''
        body = json.dumps(yaml.safe_load(data)) if data else '' #XXX

        assert settings.API_KEY_ID, "No Key ID"
        assert settings.API_SECRET_KEY, "No Secret key"
        #print "METHOD:", method.upper()
        string_to_sign = "%s\n%s\n%s\n%s\n%s" % (method.upper(), time_iso8601, request_uri, query_string, body)
        #string_to_sign = "%s\n%s\n%s\n%s\n%s" % ("GET", time_iso8601, request_uri, query_string, "")
        if settings.debug_mode:
            click.echo("API HOST: %s" % api_host)
            click.echo()
            click.echo("stringToSign:")
            click.echo(string_to_sign)

        digest = hmac.new(settings.API_SECRET_KEY, string_to_sign, hashlib.sha256).digest()
        signature = binascii.b2a_base64(digest).strip()

        headers = dict()
        headers['Content-Type'] = 'application/json; charset=utf-8'
        headers['X-Scalr-Key-Id'] = settings.API_KEY_ID
        headers['X-Scalr-Date'] = time_iso8601
        headers['X-Scalr-Signature'] = "%s %s" % (settings.SIGNATURE_VERSION, signature)
        #if hasattr(settings, "API_DEBUG") and settings.API_DEBUG:
        headers['X-Scalr-Debug'] = 1

        url = urlparse.urlunsplit((scheme, api_host, request_uri, '', ''))

        if settings.debug_mode:
            #click.echo("%s URL: %s" % (method.upper(), url))
            click.echo("Headers: %s " % json.dumps(headers, indent=2))
            click.echo()

        #session = requests.sessions.Session()
        #session.headers.pop('Connection')
        r = requests.request(method.lower(), url, data=body, params=payload, headers=headers)
        #session.close()
        result = r.text

    except (Exception, BaseException), e:
        if settings.debug_mode:
            raise
        raise click.ClickException(str(e))
    return result



import os

from distutils.core import setup
from distutils.command.install import install


try:
    from post_setup import main as post_install
except ImportError:
    post_install = lambda: None


class _install(install):
    def run(self):
        install.run(self)
        post_install()

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

description = "Scalr-ctl is a command-line interface to your Scalr account"


if __name__ == '__main__':

    setup(
        name='scalr-ctl',
        version = read("VERSION").strip(),
        description = description,
        author = "Scalr Inc.",
        author_email = "dmitry@scalr.com",
        url = "https://scalr.net",
        license = "GPL",
        platforms = "any",
        long_description=read('README'),
        classifiers=[
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'License :: OSI Approved :: GNU General Public License (GPL)',
            'Operating System :: OS Independent',
            'Development Status :: 4 - Beta',
            'Environment :: Console',
            'Intended Audience :: System Administrators',
            'Topic :: Utilities'
            ],
        packages = [
            "scalrctl",
            "scalrctl.commands",
        ],
        include_package_data=True,
        install_requires=[
            'Click',
            'prettytable',
            'pyyaml',
            'requests'
        ],
        entry_points='''
            [console_scripts]
            scalr-ctl=scalrctl.app:cli
        ''',
    )


__author__ = 'Dmitriy Korsakov'

import click

import re
import json
import traceback
from collections import defaultdict

from scalrctl import settings
from scalrctl import request
from scalrctl import spec
from scalrctl.view import build_table, build_tree


enabled = False


class SubCommandMeta(type):

    def __new__(mcs, name, parents, dct):
        route = dct.get("route")
        c = super(SubCommandMeta, mcs).__new__(mcs, name, parents, dct)
        if name!="SubCommand" and route:
            SubCommand._siblings[route].append(c)
        return c


class SubCommand(object):
    __metaclass__ = SubCommandMeta
    _siblings = defaultdict(list)
    name = None
    route = None
    method = None
    enabled = False
    _table_columns = None
    api_level = "user"

    mutable_body_parts = None # temporary, object definitions in YAML spec are not always correct
    object_reference = None # optional, e.g. '#/definitions/GlobalVariable'
    prompt_for = None #optional, Some values like GCE imageId cannot be passed through command line

    def __init__(self):
        self._table_columns = []

    @property
    def spc(self):
        return spec.Spec(spec.get_raw_spec(self.api_level), self.route, self.method)

    def get_siblings(self):
        return SubCommand._siblings.get(self.route)

    @property
    def _basepath_uri(self):
        return spec.get_raw_spec(self.api_level)["basePath"]

    @property
    def _request_template(self):
        return "%s%s" % (self._basepath_uri, self.route)

    def modify_options(self, options):
        """
        this is the place where command line options can be fixed
        after they are loaded from yaml spec
        """
        #print "In SubCommand modifier"
        for option in options:
            if self.prompt_for and option.name in self.prompt_for:
                option.prompt = option.name

            if option.name == "envId" and settings.envId:
                option.required = False

        return options


    def pre(self, *args, **kwargs):
        """
        before request is made
        """
        raw_columns = kwargs.pop("columns", False)
        if raw_columns:
            self._table_columns = raw_columns.split(",")

        settings.debug_mode = kwargs.pop("debug", False)
        settings.view = kwargs.pop("transformation", "tree")
        settings.colored_output = not kwargs.pop("nocolor", False)

        raw_filters = kwargs.pop("filters", None)

        if raw_filters:
            for pair in raw_filters.split(","):
                kv = pair.split("=")
                if 2==len(kv):
                    kwargs[kv[0]]=kv[1]


        stdin = kwargs.pop("stdin", False)

        if self.method.upper() in ("PATCH", "POST"):
            #prompting for body and then validating it
            for param in self._post_params():
                name = param["name"]

                if stdin:
                    raw = click.termui.prompt("Input %s in JSON format" % name)
                else:
                    text = ''
                    if self.method.upper() == "PATCH":
                        try:
                            for cls in self.get_siblings():
                                if cls.method.upper() == "GET":
                                    rawtext = cls().run(*args, **kwargs)

                                    if settings.debug_mode:
                                        click.echo(rawtext)

                                    json_text = json.loads(rawtext)
                                    filtered = self._filter_json_object(json_text['data'], filter_createonly=True)
                                    text = json.dumps(filtered)
                        except (Exception, BaseException), e:
                            if settings.debug_mode:
                                click.echo(traceback.format_exc())
                            else:
                                click.echo(e)
                    raw = click.edit(text)

                try:
                    user_object = json.loads(raw)
                except (Exception, BaseException), e:
                    if settings.debug_mode:
                        raise
                    raise click.ClickException(str(e))

                valid_object = self._filter_json_object(user_object)
                valid_object_str = json.dumps(valid_object)
                kwargs[name] = valid_object_str
        return args, kwargs


    def post(self, response):
        """
        after request is made
        """
        return response


    def run(self, *args, **kwargs):
        """
        callback for click subcommand
        """
        args, kwargs = self.pre(*args, **kwargs)
        uri = self._request_template
        payload = {}
        data = None

        if settings.envId and '{envId}' in uri and ('envId' not in kwargs or not kwargs['envId']):
            kwargs['envId'] = settings.envId  # XXX

        if kwargs:
            uri = self._request_template.format(**kwargs)

            for key, value in kwargs.items():
                t = "{%s}" % key
                # filtering in-body and empty params
                if value and t not in self._request_template:
                    if self.method.upper() in ("GET", "DELETE"):
                        payload[key] = value
                    elif self.method.upper() in ("POST", "PATCH"):
                        data = value  #XXX

        raw_response = request.request(self.method, uri, payload, data)
        response = self.post(raw_response)

        if settings.view == "raw":
            click.echo(raw_response)

        if raw_response:

            response_json = json.loads(response)

            if "errors" in response_json and response_json["errors"]:
                raise click.ClickException(response_json["errors"][0]['message'])

            data = response_json["data"]
            text = json.dumps(data)

            if settings.debug_mode:
                click.echo(response_json["meta"])

            if settings.view == "tree":
                click.echo(build_tree(text))

            elif settings.view == "table":
                columns = self._table_columns or self.spc.get_column_names()
                rows = []
                for block in data:
                    row = []
                    for name in columns:
                        for item in block:
                            if name.lower() == item.lower():
                                row.append(block[item])
                                break
                        else:
                            raise click.ClickException("Cannot apply filter. No such column: %s" % name)
                    if row:
                        rows.append(row)

                pagination = response_json.get("pagination", None)
                if pagination:
                    pagenum_last, current_pagenum = 1, 1
                    url_last = pagination.get('last', None)
                    if url_last:
                        number = re.search("pageNum=(\d*)", url_last)
                        pagenum_last = number.group(1) if number else 1

                    url_next = pagination.get('next', None)
                    if url_next:
                        num = re.search("pageNum=(\d*)", url_next)
                        pagenum_next = num.group(1) if num else 1
                        current_pagenum = int(pagenum_next) - 1



                click.echo(build_table(columns, rows, "Page: %s of %s" % (current_pagenum, pagenum_last))) #XXX

        return response


    def _list_mutable_body_parts(self):
        """
        finds object in yaml spec and determines it's mutable fields
        to filter user JSON
        XXX: Move to Spec (?) after TODO#3
        """
        mutable = []
        reference_path = None

        ### spcobj = spec.Spec(settings.spec, self.route, self.method)
        ### print "descr:", spcobj._result_descr

        if not self.object_reference:
            for param in spec.get_raw_spec(self.api_level)["paths"][self.route][self.method]["parameters"]:
                name = param["name"] # image
                if "schema" in param:
                    reference_path = param["schema"]['$ref'] # #/definitions/Image
                elif 'readOnly' not in param:
                    # e.g. POST /{envId}/farms/{farmId}/actions/terminate/
                    mutable.append(name)
        else:
            #XXX: Temporary code, see GlobalVariableDetailEnvelope or "role-global-variables update"
            reference_path = self.object_reference
        #print "reference_path", reference_path
        
        if reference_path:
            parts = reference_path.split("/")
            object =  spec.get_raw_spec(self.api_level)[parts[1]][parts[2]]
            object_properties = object["properties"]
            for property, descr in object_properties.items():
                if 'readOnly' not in descr or not descr['readOnly']:
                        mutable.append(property)
        return mutable

    def _list_createonly_properties(self):
        """
        Some properties (mostly IDs) cannot be marked as read-only because they are
        passed in PUT-methods (but still must be filtered in UPDATE-methods)
        """
        result = []
        properties = self.spc._result_descr['properties']
        data = properties['data']
        if 'items' in data:
            path = data['items']['$ref']
        else:
            path = data['$ref']
        obj = self.spc.lookup(path)
        if 'x-createOnly' in obj:
            result = obj['x-createOnly']
        return result

    def _filter_json_object(self, obj, filter_createonly=False):
        """
        removes immutable parts from JSON object before sending it in POST or PATCH
        """
        #XXX: make it recursive
        result = {}
        mutable_parts = self.mutable_body_parts or self._list_mutable_body_parts()
        for name, value in obj.items():
            if filter_createonly and name in self._list_createonly_properties():
                continue
            elif name in mutable_parts:
                result[name] = obj[name]
        return result

    def _post_params(self):
        """
        Determines list of body params
        e.g. 'image' JSON object for 'change-attributes' command
        """
        params = []
        m = spec.get_raw_spec(self.api_level)["paths"][self.route][self.method]
        if "parameters" in m:
            for parameter in m['parameters']:
                params.append(parameter)
        return params


__author__ = 'Dmitriy Korsakov'
__doc__ = 'Script management'


from scalrctl import commands


name = "script"
enabled = True


def callback(*args, **kwargs):
    """
    print('in script module')
    print(args)
    print(kwargs)
    """
    pass


class Script(commands.SubCommand):
    pass


class ListScripts(Script):
    name = "list"
    route = "/{envId}/scripts/"
    method = "get"
    enabled = True


class RetrieveScript(Script):
    name = "retrieve"
    route = "/{envId}/scripts/{scriptId}/"
    method = "get"
    enabled = True


class CreateScript(Script):
    name = "create"
    route = "/{envId}/scripts/"
    method = "post"
    enabled = True


class ChangeScriptAttrs(Script):
    name = "change-attributes"
    route = "/{envId}/scripts/{scriptId}/"
    method = "patch"
    enabled = True


class DeleteScript(Script):
    name = "delete"
    route = "/{envId}/scripts/{scriptId}/"
    method = "delete"
    enabled = True
